# This cell computes 16 cycles that represent a basis in Hochschild-cohomology of the genus 2 algebra
load "PMCAlgHoch.sage"
import time

my_pmc = antipodal_matching(2)
print 'Computing diffs'
time_1 = time.clock()
diffs = hoch_coh(my_pmc)
time_2 = time.clock()
print '(took '+str(time_2 - time_1)+').  Computing M'
M = diffs_to_mx(diffs)
time_3 = time.clock()
print '(took '+str(time_3 - time_2)+').  Computing kernel'
Z = M.kernel()
time_4 = time.clock()
print '(took '+str(time_4 - time_3)+').  Computing image'
B = M.image()
time_5 = time.clock()
print '(took '+str(time_5 - time_4)+').  Computing kernel/image'
H = Z/B
time_6 = time.clock()
print '(took '+str(time_6 - time_5)+').  Computing basis'
Hbasis_vector = [H.lift(x) for x in H.basis()]
time_7 = time.clock()
print '(took '+str(time_7 - time_6)+'). Tracing basis back to diffs.keys elements'
Hbasis_indices = [[i for i in range(len(h)) if h[i] == 1] for h in Hbasis_vector]
Hbasis = [[diffs.keys()[i] for i in v] for v in Hbasis_indices]
time_8 = time.clock()
print '(took '+str(time_8 - time_7)+').  Total time: '+str(time_8 - time_1)


# This cell initializes the cycles found in the previous cell into KoszulBasisElt form, 
# and separates them by degree.  (So HH_genus_two[-i] is HH^i(genus two algebra))

Hbasis2 = [[KoszulBasisElt(x,my_pmc,progress = True) for x in h] for h in Hbasis]

HH_genus_two = {0:[],-1:[],-2:[],-3:[]}
for h in Hbasis2:
    HH_genus_two[h[0].big_grading()[0]].append(h)

for i in range(4):
    print 'HH^'+str(i)+' of the genus 2, spinc 0 algebra is rank '+str(len(HH_genus_two[-i]))

# The identity element
one = KoszulElt(HH_genus_two[0][0],progress = True)

# Visual approximation of what we are about to do:
#
#     e0 
# d(one)        e1
#          d2(one)      e2
#                  d3(one)      e3
#                          d4(one)      e4
#                                  d5(one)      e5
#                                          d6(one)
#
# 
# Each element in the lower diagonal is the Hochschild 
# differential of the element above it, and one+tau of 
# the element to the right of it.  We will write
# d2(one) =: de1, d3(one) =: de2 and so on. 

e0 = one.square()
de0 = one.square().differential()
e1 = de0.one_plus_tau_lift()
de1 = e1.differential()
e2prime = de1.one_plus_tau_lift()
de2prime = e2prime.differential()
e3prime = de2prime.one_plus_tau_lift()
de3prime = e3prime.differential()
# Now we find that de3prime = d4(one) is not in the image of one plus tau

print de3prime.is_one_plus_tau()

squares_in_de3prime = KoszulElt([x for x in de3prime.data if x == x.tau()])
print "There are "+str(len(squares_in_de3prime.data))+" squares in de3prime."

# Because of this, we have to subtract off something in the image 
# of an earlier differential in the hv spectral sequence.  This
# is what we're doing in this cell

# We will write squares_in_de3prime as the square of something of the form differential(-).
# Recall that the matrix M encodes the differential, so we can
# find a lift using M.solve_left().  First we need to encode squares_in_de3 as
# a vector.

# This part is just to speed up the computation of the vector.
diff_key_hashes = []
for i in range(len(diffs.keys())):
    diff_key_hashes.append(hash(diffs.keys()[i]))


# We'll convert squares_in_de3prime to the vector v_sparse
v_indices = []
for y in squares_in_de3prime.data:
    v_indices.append(diff_key_hashes.index( hash( (y.data[0],y.data[1]) ) ) )
v_dense = vector(GF(2),len(diffs.keys()))
for i in v_indices:
    v_dense[i] = 1
v_sparse =v_dense.sparse_vector()


# This next line would fail if d4(one) = de3 was nonzero in HH^2
v_lift = M.solve_left(v_sparse)

# Now we'll convert v_lift into a KoszulElt and square it.  This is the element x we add to e_2.
x = KoszulElt([KoszulBasisElt(diffs.keys()[i],my_pmc,progress = True).square() for i in range(len(v_lift)) if v_lift[i] == 1])

# We correct e2, de2, e3, de3 using this element.
e2 = e2prime + x
de2 = de2prime + x.differential()
e3= de2.one_plus_tau_lift()
de3 = e3.differential()

# Now we check whether de3 can be lifted by one plus tau
if de3.is_one_plus_tau():
    print "de3 can be lifted by one plus tau."

# Then keep going
de3.progress = True
e4prime = de3.one_plus_tau_lift()
de4prime = e4prime.differential()
e5prime = de4prime.one_plus_tau_lift()
de5prime = e5prime.differential()

# We see de5prime doesn't vanish on the nose.
# We make a computation similar to a few cells ago to see that it is a boundary.
squares_in_de5prime = [x for x in de5prime.data if x == x.tau()]
w_indices = []
for y in squares_in_de5prime:
    w_indices.append(diff_key_hashes.index( hash( (y.data[0],y.data[1]) ) ) )
w_dense = vector(GF(2),len(diffs.keys()))
for i in w_indices:
    w_dense[i] = 1
w_sparse = w_dense.sparse_vector()

# This part would fail if de5prime were not a boundary.
w_lift = M.solve_left(w_sparse)

y = KoszulElt([KoszulBasisElt(diffs.keys()[i],my_pmc,progress = True).square() for i in range(len(w_lift)) if w_lift[i] == 1])
e4 = e4prime + y
de4 = de4prime + y.differential()
e5 = de4.one_plus_tau_lift()
de5 = e5.differential()
if de5.is_one_plus_tau():
    print "de5.is_one_plus_tau() returns true. So, d^6(1|1|1|1) does, indeed, vanish."

